# (C) 2021 GoodData Corporation
from __future__ import annotations

from typing import Optional

from gooddata_pandas import __version__
from gooddata_pandas.dataframe import DataFrameFactory
from gooddata_pandas.series import SeriesFactory
from gooddata_sdk import GoodDataSdk

USER_AGENT = f"gooddata-pandas/{__version__}"
"""Extra segment of the User-Agent header that will be appended to standard gooddata-sdk user agent."""


class GoodPandas:
    """
    Facade to access factories that create pandas Series and DataFrames using analytics computed by GoodData.CN.
    """

    def __init__(
        self,
        host: str,
        token: str,
        headers_host: Optional[str] = None,
        **custom_headers_: Optional[str],
    ) -> None:
        if headers_host is not None:
            custom_headers_["Host"] = headers_host
        self._sdk = GoodDataSdk.create(host, token, USER_AGENT, **custom_headers_)
        self._series_per_ws: dict[str, SeriesFactory] = dict()
        self._frames_per_ws: dict[str, DataFrameFactory] = dict()

    @property
    def sdk(self) -> GoodDataSdk:
        return self._sdk

    def series(self, workspace_id: str) -> SeriesFactory:
        """
        Creates factory to use for construction of pandas.Series.

        :param workspace_id: workspace to which the factory will be bound
        :return: always one same instance for given workspace
        """
        if workspace_id not in self._series_per_ws:
            self._series_per_ws[workspace_id] = SeriesFactory(sdk=self._sdk, workspace_id=workspace_id)

        return self._series_per_ws[workspace_id]

    def data_frames(self, workspace_id: str) -> DataFrameFactory:
        """
        Creates factory to use for construction of pandas.DataFrame.

        :param workspace_id: workspace to which the factory will be bound
        :return: always one same instance for given workspace
        """
        if workspace_id not in self._frames_per_ws:
            self._frames_per_ws[workspace_id] = DataFrameFactory(sdk=self._sdk, workspace_id=workspace_id)

        return self._frames_per_ws[workspace_id]



# (C) 2021 GoodData Corporation
try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata  # type: ignore # mypy issue #1153

__version__: str = metadata.version("gooddata-pandas")



# (C) 2022 GoodData Corporation
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import pandas
from attrs import define, field, frozen

from gooddata_sdk import BareExecutionResponse, ExecutionResult, ResultCacheMetadata, ResultSizeDimensions

_DEFAULT_PAGE_SIZE = 100
_DataHeaders = List[List[Any]]
_DataArray = List[Union[int, None]]
LabelOverrides = Dict[str, Dict[str, Dict[str, str]]]


@frozen
class _DataWithHeaders:
    data: List[_DataArray]
    """extracted data; either array of values for one-dimensional result or array of arrays of values"""

    data_headers: Tuple[_DataHeaders, Optional[_DataHeaders]]
    """per-dimension headers for the data"""

    grand_totals: Tuple[Optional[List[_DataArray]], Optional[List[_DataArray]]]
    """per-dimension grand total data"""

    grand_total_headers: Tuple[Optional[_DataHeaders], Optional[_DataHeaders]]
    """per-dimension grand total headers"""


@define
class _AccumulatedData:
    """
    Utility class to offload code from the function that extracts all data and headers for a
    particular paged result. The method drives the paging and calls out to this class to accumulate
    the essential data and headers from the page.
    """

    data: List[_DataArray] = field(init=False, factory=list)
    data_headers: List[Optional[_DataHeaders]] = field(init=False, factory=lambda: [None, None])
    grand_totals: List[Optional[List[_DataArray]]] = field(init=False, factory=lambda: [None, None])
    grand_totals_headers: List[Optional[_DataHeaders]] = field(init=False, factory=lambda: [None, None])

    def accumulate_data(self, from_result: ExecutionResult) -> None:
        """
        if one-dimensional result, the data is single array, so this adds the elements of that array into 'data'
        if two-dimensional, the data is array of arrays, so this adds as many arrays as there are table rows
        """
        self.data.extend(from_result.data)

    def extend_existing_row_data(self, from_result: ExecutionResult) -> None:
        offset = from_result.paging_offset[0]

        for i in range(len(from_result.data)):
            self.data[offset + i].extend(from_result.data[i])

    def accumulate_headers(self, from_result: ExecutionResult, from_dim: int) -> None:
        """
        Accumulates headers for particular dimension of a result into the provided `data_headers` array at the index
        matching the dimension index. This will mutate the `data_headers`
        """
        if self.data_headers[from_dim] is None:
            self.data_headers[from_dim] = from_result.get_all_headers(dim=from_dim)
        else:
            for idx, headers in enumerate(from_result.get_all_headers(dim=from_dim)):
                cast(_DataHeaders, self.data_headers[from_dim])[idx].extend(headers)

    def accumulate_grand_totals(
        self, from_result: ExecutionResult, paging_dim: int, response: BareExecutionResponse
    ) -> None:
        """
        accumulates grand totals from the results; processes all grand totals on all dimensions; the method
        needs to know in which direction is the paging happening so that it can append new grand total data.
        """
        grand_totals = from_result.grand_totals
        if not len(grand_totals):
            return

        # get dimension indexes mapping from response like {"dim1": 0, "dim0": 1}
        dim_idx_dict = {dim["localIdentifier"]: idx for idx, dim in enumerate(response.dimensions)}

        for grand_total in grand_totals:
            # 2-dim results have always 1-dim grand totals (3-dim results have 2-dim gt but DataFrame stores 2D only)
            dims = grand_total["totalDimensions"]
            assert len(dims) == 1, "Only 2-dimensional results are supported"
            dim_idx = dim_idx_dict[dims[0]]
            # the dimension id specified on the grand total says from what dimension were
            # the grand totals calculated (1 for column totals or 0 for row totals);
            #
            # the grand totals themselves should, however be placed in the opposite dimension:
            #
            # column totals are extra rows at the end of the data
            # row totals are extra columns at the right 'edge' of the data
            opposite_dim = 1 if dim_idx == 0 else 0

            if self.grand_totals[opposite_dim] is None:
                # grand totals not initialized yet; initialize both data and headers by making
                # a shallow copy from the results
                self.grand_totals[opposite_dim] = grand_total["data"][:]
                # TODO: row total measure headers are currently not supported (only aggregation info w/o measure label)
                #       measure header defs are under ["headerGroups"][>0]
                self.grand_totals_headers[opposite_dim] = grand_total["dimensionHeaders"][0]["headerGroups"][0][
                    "headers"
                ][:]
            elif paging_dim != opposite_dim:
                # grand totals are already initialized and the code is paging in the direction that reveals
                # additional grand total values; append them accordingly; no need to consider total headers:
                # that is because only the grand total data is subject to paging
                grand_totals_item = cast(List[_DataArray], self.grand_totals[opposite_dim])
                if opposite_dim == 0:
                    # have column totals and paging 'to the right'; totals for the new columns are revealed so
                    # extend existing data arrays
                    for total_idx, total_data in enumerate(grand_total["data"]):
                        grand_totals_item[total_idx].extend(total_data)
                else:
                    # have row totals and paging down, keep adding extra rows
                    grand_totals_item.extend(grand_total["data"])

    def result(self) -> _DataWithHeaders:
        return _DataWithHeaders(
            data=self.data,
            data_headers=(cast(_DataHeaders, self.data_headers[0]), self.data_headers[1]),
            grand_totals=(self.grand_totals[0], self.grand_totals[1]),
            grand_total_headers=(self.grand_totals_headers[0], self.grand_totals_headers[1]),
        )


@define
class DataFrameMetadata:
    # row line location where total header is located per index header column
    # example:
    # Category  | Country | Budget
    # Car       | Arizona |    100
    #           | Texas   |     50
    #           | SUM     |    150
    # Train     | Arizona |    200    =>    [[7],[3,6]]
    #           | Texas   |    100
    #           | AVG     |    150
    # SUM       |         |    450
    row_totals_indexes: List[List[int]]

    execution_response: BareExecutionResponse

    @classmethod
    def from_data(
        cls,
        headers: Tuple[_DataHeaders, Optional[_DataHeaders]],
        execution_response: BareExecutionResponse,
    ) -> "DataFrameMetadata":
        row_totals_indexes = [
            [idx for idx, hdr in enumerate(dim) if hdr is not None and "totalHeader" in hdr] for dim in headers[0]
        ]
        return cls(
            row_totals_indexes=row_totals_indexes,
            execution_response=execution_response,
        )


def _read_complete_execution_result(
    execution_response: BareExecutionResponse,
    result_cache_metadata: ResultCacheMetadata,
    result_size_dimensions_limits: ResultSizeDimensions,
    result_size_bytes_limit: Optional[int] = None,
    page_size: int = _DEFAULT_PAGE_SIZE,
) -> _DataWithHeaders:
    """
    Extracts all data and headers for an execution result. This does page around the execution result to extract
    everything from the paged API.

    :param execution_response: execution response to work with;
    :return:
    """
    num_dims = len(execution_response.dimensions)
    offset = [0] * num_dims
    limit = [page_size] * num_dims
    acc = _AccumulatedData()

    result_size_limits_checked = False

    while True:
        # top-level loop pages through the first dimension;
        #
        # if one-dimensional result, it pages over an array of data
        # if two-dimensional result, it pages over table rows
        result = execution_response.read_result(offset=offset, limit=limit)

        if not result_size_limits_checked:
            result.check_dimensions_size_limits(result_size_dimensions_limits)
            result_cache_metadata.check_bytes_size_limit(result_size_bytes_limit)
            result_size_limits_checked = True

        acc.accumulate_data(from_result=result)
        acc.accumulate_headers(from_result=result, from_dim=0)
        acc.accumulate_grand_totals(from_result=result, paging_dim=0, response=execution_response)

        if num_dims > 1:
            # when result is two-dimensional make sure to read the column headers and column totals
            # just once - when scrolling 'to the right' for the first time;
            load_headers_and_totals = False
            if acc.data_headers[1] is None:
                acc.accumulate_headers(from_result=result, from_dim=1)
                load_headers_and_totals = True

            if not result.is_complete(dim=1):
                # have two-dimensional result (typical table) and the page does not contain
                # all the columns.
                #
                # page 'to the right' to get data from all columns, extend existing rows with that data
                offset = [offset[0], result.next_page_start(dim=1)]
                while True:
                    result = execution_response.read_result(offset=offset, limit=limit)
                    acc.extend_existing_row_data(from_result=result)

                    if load_headers_and_totals:
                        acc.accumulate_headers(from_result=result, from_dim=1)
                        acc.accumulate_grand_totals(from_result=result, paging_dim=1, response=execution_response)

                    if result.is_complete(dim=1):
                        break

                    offset = [offset[0], result.next_page_start(dim=1)]

        if result.is_complete(dim=0):
            break

        offset = [result.next_page_start(dim=0), 0] if num_dims > 1 else [result.next_page_start(dim=0)]

    return acc.result()


def _create_header_mapper(
    response: BareExecutionResponse,
    dim: int,
    label_overrides: Optional[LabelOverrides] = None,
    use_local_ids_in_headers: bool = False,
) -> Callable[[Any, Optional[int]], Optional[str]]:
    """
    Prepares header mapper function which is doing header structures translations into appropriate label used
    in a dataframe
    :param response: Response structure to gather dimension header details
    :param dim: dimension id
    :param label_overrides: label overrides
    :param use_local_ids_in_headers: Use local identifiers of header attributes and metrics. Optional.

    :return: Mapper function
    """
    if label_overrides is None:
        label_overrides = {}

    dim_descriptor = response.dimensions[dim]
    attribute_labels = label_overrides.get("labels", {})
    measure_labels = label_overrides.get("metrics", {})

    def _mapper(header: Any, header_idx: Optional[int]) -> Optional[str]:
        label = None
        if header is None:
            pass
        elif "attributeHeader" in header:
            if "labelValue" in header["attributeHeader"]:
                label = header["attributeHeader"]["labelValue"]
            elif "labelName" in header["attributeHeader"]:
                attr_local_id = header["attributeHeader"]["localIdentifier"]
                if use_local_ids_in_headers:
                    label = attr_local_id
                else:
                    if attr_local_id in attribute_labels:
                        label = attribute_labels[attr_local_id]["title"]
                    else:
                        label = header["attributeHeader"]["labelName"]
        elif "measureHeader" in header and header_idx is not None:
            measure_idx = header["measureHeader"]["measureIndex"]
            measure_descriptor = dim_descriptor["headers"][header_idx]["measureGroupHeaders"][measure_idx]

            if use_local_ids_in_headers:
                label = measure_descriptor["localIdentifier"]
            else:
                if measure_descriptor["localIdentifier"] in measure_labels:
                    label = measure_labels[measure_descriptor["localIdentifier"]]["title"]
                elif "name" in measure_descriptor:
                    label = measure_descriptor["name"]
                else:
                    label = measure_descriptor["localIdentifier"]
        elif "totalHeader" in header:
            label = header["totalHeader"]["function"]
        return label

    return _mapper


def _headers_to_index(
    dim_idx: int,
    headers: Tuple[_DataHeaders, Optional[_DataHeaders]],
    response: BareExecutionResponse,
    label_overrides: LabelOverrides,
    use_local_ids_in_headers: bool = False,
) -> Optional[pandas.Index]:
    if len(response.dimensions) <= dim_idx or not len(response.dimensions[dim_idx]["headers"]):
        return None

    mapper = _create_header_mapper(
        response=response,
        dim=dim_idx,
        label_overrides=label_overrides,
        use_local_ids_in_headers=use_local_ids_in_headers,
    )

    return pandas.MultiIndex.from_arrays(
        [
            tuple(mapper(header, header_idx) for header in header_group)
            for header_idx, header_group in enumerate(cast(_DataHeaders, headers[dim_idx]))
        ],
        names=[mapper(dim_header, None) for dim_header in (response.dimensions[dim_idx]["headers"])],
    )


def _merge_grand_totals_into_data(extract: _DataWithHeaders) -> Union[_DataArray, List[_DataArray]]:
    """
    Merges grand totals into the extracted data. this function will mutate the extracted data, extending
    the rows and columns with grand totals. Going with mutation here so as not to copy arrays around
    """
    data: List[_DataArray] = extract.data

    if extract.grand_totals[0] is not None:
        # column totals are computed into extra rows, one row per column total
        # add those rows at the end of the data rows
        data.extend(extract.grand_totals[0])

    if extract.grand_totals[1] is not None:
        # row totals are computed into extra columns that should be appended to
        # existing data rows
        for row_idx, cols_to_append in enumerate(extract.grand_totals[1]):
            data[row_idx].extend(cols_to_append)

    return data


def _merge_grand_total_headers_into_headers(extract: _DataWithHeaders) -> Tuple[_DataHeaders, Optional[_DataHeaders]]:
    """
    Merges grand total headers into data headers. This function will mutate the extracted data.
    """
    headers: Tuple[_DataHeaders, Optional[_DataHeaders]] = extract.data_headers

    for dim_idx, grand_total_headers in enumerate(extract.grand_total_headers):
        if grand_total_headers is None:
            continue
        header = cast(List[List[Any]], headers[dim_idx])
        header[0].extend(grand_total_headers)
        padding = [None] * len(grand_total_headers)
        for other_headers in header[1:]:
            other_headers.extend(padding)

    return headers


def convert_execution_response_to_dataframe(
    execution_response: BareExecutionResponse,
    result_cache_metadata: ResultCacheMetadata,
    label_overrides: LabelOverrides,
    result_size_dimensions_limits: ResultSizeDimensions,
    result_size_bytes_limit: Optional[int] = None,
    use_local_ids_in_headers: bool = False,
) -> Tuple[pandas.DataFrame, DataFrameMetadata]:
    """
    Converts execution result to a pandas dataframe, maintaining the dimensionality of the result.

    Because the result itself does not contain all the necessary metadata to do the full conversion, this method
    expects that the execution _response_.

    :param label_overrides: label overrides
    :param execution_response: execution response through which the result can be read and converted to a dataframe
    :return: a new dataframe
    """
    extract = _read_complete_execution_result(
        execution_response=execution_response,
        result_cache_metadata=result_cache_metadata,
        result_size_dimensions_limits=result_size_dimensions_limits,
        result_size_bytes_limit=result_size_bytes_limit,
    )
    full_data = _merge_grand_totals_into_data(extract)
    full_headers = _merge_grand_total_headers_into_headers(extract)

    df = pandas.DataFrame(
        data=full_data,
        index=_headers_to_index(
            dim_idx=0,
            headers=full_headers,
            response=execution_response,
            label_overrides=label_overrides,
            use_local_ids_in_headers=use_local_ids_in_headers,
        ),
        columns=_headers_to_index(
            dim_idx=1,
            headers=full_headers,
            response=execution_response,
            label_overrides=label_overrides,
            use_local_ids_in_headers=use_local_ids_in_headers,
        ),
    )

    return df, DataFrameMetadata.from_data(headers=full_headers, execution_response=execution_response)



# (C) 2021 GoodData Corporation
from __future__ import annotations

from typing import Any, Optional, Union

from gooddata_pandas.utils import (
    ColumnsDef,
    IndexDef,
    LabelItemDef,
    _str_to_obj_id,
    _to_attribute,
    _to_item,
    _typed_attribute_value,
)
from gooddata_sdk import (
    Attribute,
    AttributeFilter,
    CatalogWorkspaceContent,
    ExecutionDefinition,
    ExecutionResponse,
    Filter,
    GoodDataSdk,
    Metric,
    MetricValueFilter,
    ObjId,
)


class ExecutionDefinitionBuilder:
    _DEFAULT_INDEX_NAME: str = "0"

    def __init__(self, columns: ColumnsDef, index_by: Optional[IndexDef] = None) -> None:
        self._attributes: list[Attribute] = []
        self._metrics: list[Metric] = []
        self._col_to_attr_idx: dict[str, int] = dict()
        self._index_to_attr_idx: dict[str, int] = dict()
        self._col_to_metric_idx: dict[str, int] = dict()

        self._process_columns(columns)
        self._process_index(index_by)

    @property
    def col_to_attr_idx(self) -> dict[str, int]:
        return self._col_to_attr_idx

    @property
    def index_to_attr_idx(self) -> dict[str, int]:
        return self._index_to_attr_idx

    @property
    def col_to_metric_idx(self) -> dict[str, int]:
        return self._col_to_metric_idx

    def _process_columns(self, columns: ColumnsDef) -> None:
        for column_name, column_def in columns.items():
            self._add_column(column_name, column_def)

    def _add_column(self, column_name: str, column_def: Union[str, Attribute, Metric, ObjId]) -> None:
        item = _to_item(column_def)
        if isinstance(item, Attribute):
            # prevent double-add for attributes that are using same labels. this has no real effect on the result
            # apart from extra load/size of the result that would contain the duplicates.
            #
            # this may typically happen when same labels are used for indexing & data
            attr_index = self._find_attribute_index(item)

            if attr_index is not None:
                self._col_to_attr_idx[column_name] = attr_index
            else:
                self._col_to_attr_idx[column_name] = len(self._attributes)
                self._attributes.append(item)
        elif isinstance(item, Metric):
            self._col_to_metric_idx[column_name] = len(self._metrics)
            self._metrics.append(item)

    def _process_index(self, index_by: Optional[IndexDef] = None) -> None:
        if index_by is None:
            return

        if not isinstance(index_by, dict):
            _index_by = {self._DEFAULT_INDEX_NAME: index_by}
        else:
            _index_by = index_by

        for index_name, index_def in _index_by.items():
            if isinstance(index_def, str) and (index_def in self._col_to_attr_idx):
                # known attribute defined in columns referenced by the column key
                attr_index = self._col_to_attr_idx[index_def]
                self._index_to_attr_idx[index_name] = attr_index
            elif isinstance(index_def, str) and (index_def in self._col_to_metric_idx):
                # known metric defined in columns referenced by the column key - index_by cannot be a metric
                raise ValueError(f"Invalid index_col item {index_def}, type={type(self._col_to_metric_idx[index_def])}")
            else:
                self._process_index_item_without_col_ref(index_name, index_def)

    def _find_attribute_index(self, item: Attribute) -> Union[int | None]:
        existing_attr_idx = next(
            (idx for idx, attr in enumerate(self._attributes) if item.has_same_label(attr)),
            None,
        )
        return existing_attr_idx

    def _process_index_item_without_col_ref(self, index_name: str, index_def: LabelItemDef) -> None:
        norm_index_def = index_def
        if isinstance(index_def, str) and (not index_def.startswith("label/")):
            # it is not a obj id in string form - consider it label_id and extend it to obj id string form
            norm_index_def = f"label/{index_def}"
        attr_item = _to_attribute(norm_index_def)
        attr_index = self._find_attribute_index(attr_item)

        if attr_index is not None:
            self._index_to_attr_idx[index_name] = attr_index
        else:
            self._index_to_attr_idx[index_name] = len(self._attributes)
            self._attributes.append(attr_item)

    def _update_filter_ids(self, filter_by: Optional[Union[Filter, list[Filter]]] = None) -> Optional[list[Filter]]:
        filters = [filter_by] if isinstance(filter_by, Filter) else filter_by
        if filters:
            for _filter in filters:
                if isinstance(_filter, AttributeFilter) and isinstance(_filter.label, str):
                    if _filter.label in self._col_to_attr_idx:
                        _filter.label = self._attributes[self._col_to_attr_idx[_filter.label]].label
                    elif _filter.label in self._index_to_attr_idx:
                        _filter.label = self._attributes[self._index_to_attr_idx[_filter.label]].label
                    elif _filter.label in self._col_to_metric_idx:
                        raise ValueError(f"AttributeFilter instance referencing metric [{_filter.label}]")
                    else:
                        _filter.label = _str_to_obj_id(_filter.label) or _filter.label
                elif isinstance(_filter, MetricValueFilter) and isinstance(_filter.metric, str):
                    if _filter.metric in self._col_to_metric_idx:
                        # Metric is referenced by local_id which was already generated during creation of columns
                        # When Metric filter contains ObjId reference, it does not need to be modified
                        _filter.metric = self._metrics[self._col_to_metric_idx[_filter.metric]].local_id

        return filters

    def build_execution_definition(
        self, filter_by: Optional[Union[Filter, list[Filter]]] = None
    ) -> ExecutionDefinition:
        dimensions = [
            ["measureGroup"] if len(self._metrics) else None,
            [a.local_id for a in self._attributes] if len(self._attributes) else None,
        ]

        filters = self._update_filter_ids(filter_by)

        return ExecutionDefinition(
            attributes=self._attributes,
            metrics=self._metrics,
            filters=filters,
            dimensions=dimensions,
        )


def _compute(
    sdk: GoodDataSdk,
    workspace_id: str,
    columns: ColumnsDef,
    index_by: Optional[IndexDef] = None,
    filter_by: Optional[Union[Filter, list[Filter]]] = None,
) -> tuple[ExecutionResponse, dict[str, int], dict[str, int], dict[str, int]]:
    """
    Creates execution-by-convention to retrieve data for frame with the provided columns that is optionally
    indexed by index_by label and optionally also filtered.

    The convention is as follows:

    -  If the columns contain labels and facts/metric, then two dimensional result will be created. first dim
       will contain measureGroup, second dimension will be all the attributes

    -  If indexing is desired, then the index label will be placed in the attribute dimension and will be the
       first in that dimension

    -  Otherwise the execution will be single-dim and will contain either measureGroup or all attributes

    Only columns contain always full identification of objects in a form { 'index', 'full identification' }.
    index_by and filters may contain references to columns using the index!

    The compute will return execution response and two mappings:
    -  pandas name to index where headers appear in the attribute dimension
    -  pandas col name to index where headers appear in metric dimension
    -  pandas index name to index where headers appear in the attribute dimension
    """

    builder = ExecutionDefinitionBuilder(columns, index_by)
    exec_def = builder.build_execution_definition(filter_by)

    return (
        sdk.compute.for_exec_def(workspace_id, exec_def),
        builder.col_to_attr_idx,
        builder.col_to_metric_idx,
        builder.index_to_attr_idx,
    )


_RESULT_PAGE_LEN = 1000


#
# Note: both of the extract functions assume the number of measures requested for the data frame is less than
# page limit enforced by the server. The function for getting data frame for measures only does not paging at all
# and the one that gets data from possibly two-dimensional result always pages only through the dimension that
# contains attributes.
#


def _extract_for_metrics_only(response: ExecutionResponse, cols: list, col_to_metric_idx: dict) -> dict:
    exec_def = response.exec_def
    result = response.read_result(len(exec_def.metrics))
    data = dict()

    for col in cols:
        data[col] = [result.data[col_to_metric_idx[col]]]

    return data


def _typed_result(catalog: CatalogWorkspaceContent, attribute: Attribute, result_values: list[Any]) -> list[Any]:
    """Convert result_values to proper data types."""
    catalog_attribute = catalog.find_label_attribute(attribute.label)
    if catalog_attribute is None:
        raise ValueError(f"Unable to find attribute {attribute.label} in catalog")
    return [_typed_attribute_value(catalog_attribute, value) for value in result_values]


def _extract_from_attributes_and_maybe_metrics(
    response: ExecutionResponse,
    catalog: CatalogWorkspaceContent,
    cols: list[str],
    col_to_attr_idx: dict[str, int],
    col_to_metric_idx: dict[str, int],
    index_to_attr_idx: Optional[dict[str, int]] = None,
) -> tuple[dict, dict]:

    exec_def = response.exec_def
    offset = [0 for _ in exec_def.dimensions]
    limit = [len(exec_def.metrics), _RESULT_PAGE_LEN] if exec_def.has_metrics() else [_RESULT_PAGE_LEN]
    attribute_dim = 1 if exec_def.has_metrics() else 0
    result = response.read_result(limit=limit, offset=offset)
    safe_index_to_attr_idx = index_to_attr_idx if index_to_attr_idx is not None else dict()

    # mappings from column name to Attribute
    index_to_attribute = {index_name: exec_def.attributes[i] for index_name, i in safe_index_to_attr_idx.items()}
    col_to_attribute = {col: exec_def.attributes[i] for col, i, in col_to_attr_idx.items()}

    # datastructures to return
    index: dict[str, list[Any]] = {idx_name: [] for idx_name in safe_index_to_attr_idx}
    data: dict[str, list[Any]] = {col: [] for col in cols}

    while True:
        for idx_name in index:
            rs = result.get_all_header_values(attribute_dim, safe_index_to_attr_idx[idx_name])
            attribute = index_to_attribute[idx_name]
            index[idx_name] += _typed_result(catalog, attribute, rs)
        for col in cols:
            if col in col_to_attr_idx:
                rs = result.get_all_header_values(attribute_dim, col_to_attr_idx[col])
                attribute = col_to_attribute[col]
                data[col] += _typed_result(catalog, attribute, rs)
            elif col_to_metric_idx[col] < len(result.data):
                data[col] += result.data[col_to_metric_idx[col]]
        if result.is_complete(attribute_dim):
            break

        offset[attribute_dim] = result.next_page_start(attribute_dim)
        result = response.read_result(limit=limit, offset=offset)

    return data, index


def compute_and_extract(
    sdk: GoodDataSdk,
    workspace_id: str,
    columns: ColumnsDef,
    index_by: Optional[IndexDef] = None,
    filter_by: Optional[Union[Filter, list[Filter]]] = None,
) -> tuple[dict, dict]:
    """
    Convenience function to drive computation & data extraction on behalf of the series and data frame factories.

    Given data columns and index columns, this function will create AFM execution and then read the results and
    populate data and index dicts.

    For each column in `columns`, the returned data will contain key under which there is array of data for that column
    For each index in index_by, the returned data will contain key under which there is array with data to construct the
    index. When there are multiple indexes, feed the indexes to MultiIndex.from_arrays().

    Note that as convenience it is possible to pass just single index. in that case the index dict will contain exactly
    one key of '0' (just get first value from dict when consuming the result).
    """
    result = _compute(
        sdk=sdk,
        workspace_id=workspace_id,
        index_by=index_by,
        columns=columns,
        filter_by=filter_by,
    )

    response, col_to_attr_idx, col_to_metric_idx, index_to_attr_idx = result

    exec_def = response.exec_def
    cols = list(columns.keys())

    catalog = sdk.catalog_workspace_content.get_full_catalog(workspace_id)

    if not exec_def.has_attributes():
        return _extract_for_metrics_only(response, cols, col_to_metric_idx), dict()
    else:
        return _extract_from_attributes_and_maybe_metrics(
            response,
            catalog,
            cols,
            col_to_attr_idx,
            col_to_metric_idx,
            index_to_attr_idx,
        )



# (C) 2021 GoodData Corporation
from __future__ import annotations

from typing import Optional, Tuple, Union

import pandas

from gooddata_api_client import models
from gooddata_pandas.data_access import compute_and_extract
from gooddata_pandas.result_convertor import DataFrameMetadata, LabelOverrides, convert_execution_response_to_dataframe
from gooddata_pandas.utils import (
    ColumnsDef,
    DefaultInsightColumnNaming,
    IndexDef,
    LabelItemDef,
    _to_item,
    make_pandas_index,
)
from gooddata_sdk import (
    Attribute,
    BareExecutionResponse,
    ExecutionDefinition,
    Filter,
    GoodDataSdk,
    ResultCacheMetadata,
    ResultSizeDimensions,
)


class DataFrameFactory:
    """
    Factory to create pandas.DataFrame instances.

    There are several methods in place that should provide for convenient construction of data frames:

    - ``indexed()`` - calculate measure values sliced by one or more labels, indexed by those labels
    - ``not_indexed()`` - calculate measure values sliced by one or more labels, but not indexed by those labels,
      label values will be part of the DataFrame and will be in the same row as the measure values calculated
      for them
    - ``for_items()`` - calculate measure values for a one or more items which may be labels or measures. Depending
      what items you specify, this method will create DataFrame with or without index
    - ``for_insight()`` - calculate DataFrame for insight created by GoodData.CN Analytical Designer. Depending
      on what items are in the insight, this method will create DataFrame with or without index.

    Note that all of these methods have additional levels of convenience and flexibility so their purpose is not
    limited to just what is listed above.
    """

    def __init__(self, sdk: GoodDataSdk, workspace_id: str) -> None:
        self._sdk = sdk
        self._workspace_id = workspace_id

    def indexed(
        self, index_by: IndexDef, columns: ColumnsDef, filter_by: Optional[Union[Filter, list[Filter]]] = None
    ) -> pandas.DataFrame:
        """
        Creates a data frame indexed by values of the label. The data frame columns will be created from either
        metrics or other label values.

        The computation to obtain data from GoodData.CN workspace will use all labels that you specify for both
        indexing and in columns to aggregate values of metric columns.

        Note that depending on composition of the labels, the DataFrame's index may or may not be unique.

        :param index_by: one or more labels to index by; specify either:

         - string with reference to columns key - only attribute can be referenced
         - string with id: ``some_label_id``,
         - string representation of object identifier: ``label/some_label_id``
         - object identifier: ``ObjId(id='some_label_id', type='label')``,
         - or an Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``,
         - dict containing mapping of index name to label to use for indexing - specified in one of the ways list above

        :param columns: dict mapping column name to its definition; column may be specified as:

         - object identifier: ``ObjId(id='some_id', type='<type>')`` - where type is either ``label``, ``fact``
           or ``metric``
         - string representation of object identifier: ``<type>/some_id`` - where type is either ``label``, ``fact`` or
           ``metric``
         - Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - subclass of Measure object used in the compute model: SimpleMeasure, PopDateMeasure, PopDatasetMeasure,
           ArithmeticMeasure

        :param filter_by: optional filters to apply during computation on the server, reference to filtering column
         can be one of:

         - string reference to column key or index key
         - object identifier in string form
         - object identifier: ``ObjId(id='some_label_id', type='<type>')``
         - Attribute or Metric depending on type of filter

        :return: pandas dataframe instance
        """
        data, index = compute_and_extract(
            self._sdk,
            self._workspace_id,
            columns=columns,
            index_by=index_by,
            filter_by=filter_by,
        )

        _idx = make_pandas_index(index)

        return pandas.DataFrame(data=data, index=_idx)

    def not_indexed(
        self, columns: ColumnsDef, filter_by: Optional[Union[Filter, list[Filter]]] = None
    ) -> pandas.DataFrame:
        """
        Creates a data frame with columns created from metrics and or labels.

        The computation to obtain data from GoodData.CN workspace will use all labels that you specify for both columns
        to aggregate values of metric columns.

        :param columns: dict mapping column name to its definition; column may be specified as:

         - object identifier: ``ObjId(id='some_id', type='<type>')`` - where type is either ``label``, ``fact``
           or ``metric``
         - string representation of object identifier: ``<type>/some_id`` - where type is either ``label``,
           ``fact`` or ``metric``
         - Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - subclass of Measure object used in the compute model: SimpleMeasure, PopDateMeasure, PopDatasetMeasure,
           ArithmeticMeasure

        :param filter_by: optionally specify filters to apply during computation on the server, reference to filtering
         column can be one of:

         - string reference to column key
         - object identifier in string form
         - object identifier: ``ObjId(id='some_label_id', type='<type>')``
         - Attribute or Metric depending on type of filter

        :return: pandas dataframe instance
        """

        data, _ = compute_and_extract(self._sdk, self._workspace_id, columns=columns, filter_by=filter_by)

        return pandas.DataFrame(data=data)

    def for_items(
        self, items: ColumnsDef, filter_by: Optional[Union[Filter, list[Filter]]] = None, auto_index: bool = True
    ) -> pandas.DataFrame:
        """
        Creates a data frame for a named items. This is a convenience method that will create DataFrame with or
        without index based on the context of the items that you pass.

        - If items contain labels and measures, then DataFrame with index will be created. If there is more than
          one label among the items, then hierarchical index will be created.

          You can turn this behavior using 'auto_index' parameter.

        - Otherwise DataFrame without index will be created and will contain column per item.

        You may also optionally specify filters to apply during the computation on the server.

        :param items: dict mapping item name to its definition; item may be specified as:

         - object identifier: ``ObjId(id='some_id', type='<type>')`` - where type is either ``label``, ``fact``
           or ``metric``
         - string representation of object identifier: ``<type>/some_id`` - where type is either ``label``,
           ``fact`` or ``metric``
         - Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - subclass of Measure object used in the compute model: SimpleMeasure, PopDateMeasure, PopDatasetMeasure,
           ArithmeticMeasure

        :param filter_by: optionally specify filters to apply during computation on the server, reference to filtering
         column can be one of:

         - string reference to item key
         - object identifier in string form
         - object identifier: ``ObjId(id='some_label_id', type='<type>')``
         - Attribute or Metric depending on type of filter

        :return: pandas dataframe instance
        """
        resolved_attr_cols: dict[str, LabelItemDef] = dict()
        resolved_measure_cols: ColumnsDef = dict()
        has_attributes = False
        has_measures = False

        for col_name, col_def in items.items():
            item = _to_item(col_def, local_id=col_name)

            if isinstance(item, Attribute):
                has_attributes = True
                resolved_attr_cols[col_name] = item
            else:
                has_measures = True
                resolved_measure_cols[col_name] = item

        if not auto_index or not has_measures or not has_attributes:
            columns: ColumnsDef = {**resolved_attr_cols, **resolved_measure_cols}

            return self.not_indexed(columns=columns, filter_by=filter_by)

        return self.indexed(
            index_by=resolved_attr_cols,
            columns=resolved_measure_cols,
            filter_by=filter_by,
        )

    def for_insight(self, insight_id: str, auto_index: bool = True) -> pandas.DataFrame:
        """
        Creates a data frame with columns based on the content of the insight with the provided identifier. The
        filters that are set on the insight will be applied and used for the server-side computation of the data
        for the data frame.

        This method will create DataFrame with or without index - depending on the contents of the insight. The
        rules are as follows:

        - if the insight contains both attributes and measures, it will be mapped to a DataFrame with index

          - if there are multiple attributes, hieararchical index (pandas.MultiIndex) will be used
          - otherwise a normal index will be used (pandas.Index)
          - you can use the option 'auto_index' argument to disable this logic and force no indexing

        - if the insight contains either only attributes or only measures, then DataFrame will not be indexed
          and all attribute or measures values will be used as data.

        Note that if the insight consists of single measure only, the resulting data frame is guaranteed to have
        single 'row' of data with one column per measure.

        :param insight_id: insight identifier
        :param auto_index: optionally force creation of DataFrame without index even if the data in the insight is
         eligible for indexing

        :return: pandas dataframe instance
        """
        naming = DefaultInsightColumnNaming()
        insight = self._sdk.insights.get_insight(workspace_id=self._workspace_id, insight_id=insight_id)
        filter_by = [f.as_computable() for f in insight.filters]
        columns: ColumnsDef = {
            **{naming.col_name_for_attribute(a): a.as_computable() for a in insight.attributes},
            **{naming.col_name_for_metric(m): m.as_computable() for m in insight.metrics},
        }

        return self.for_items(columns, filter_by=filter_by, auto_index=auto_index)

    def result_cache_metadata_for_exec_result_id(self, result_id: str) -> ResultCacheMetadata:
        """
        Retrieves result cache metadata for given :result_id:
        :param result_id: ID of execution result to retrieve the metadata for
        :return: corresponding result cache metadata
        """
        return self._sdk.compute.retrieve_result_cache_metadata(workspace_id=self._workspace_id, result_id=result_id)

    def for_exec_def(
        self,
        exec_def: ExecutionDefinition,
        label_overrides: Optional[LabelOverrides] = None,
        result_size_dimensions_limits: ResultSizeDimensions = (),
        result_size_bytes_limit: Optional[int] = None,
    ) -> Tuple[pandas.DataFrame, DataFrameMetadata]:
        """
        Creates a data frame using an execution definition. The data frame will respect the dimensionality
        specified in execution definition's result spec.

        Each dimension may be sliced by multiple labels. The factory will create MultiIndex for the dataframe's
        row index and the columns.

        Example of label_overrides structure:

        .. code-block:: python

            {
                "labels": {
                    "local_attribute_id": {
                        "title": "My new attribute label"
                    ,...
                },
                "metrics": {
                    "local_metric_id": {
                        "title": "My new metric label"
                    },...
                }
            }

        :param exec_def: execution definition
        :param label_overrides: label overrides for metrics and attributes
        :param result_size_dimensions_limits: A tuple containing maximum size of result dimensions. Optional.
        :param result_size_bytes_limits: Maximum size of result in bytes. Optional.
        :return: tuple holding DataFrame and DataFrame metadata
        """
        if label_overrides is None:
            label_overrides = {}

        execution = self._sdk.compute.for_exec_def(workspace_id=self._workspace_id, exec_def=exec_def)
        result_cache_metadata = self.result_cache_metadata_for_exec_result_id(execution.result_id)

        return convert_execution_response_to_dataframe(
            execution_response=execution.bare_exec_response,
            result_cache_metadata=result_cache_metadata,
            label_overrides=label_overrides,
            result_size_dimensions_limits=result_size_dimensions_limits,
            result_size_bytes_limit=result_size_bytes_limit,
        )

    def for_exec_result_id(
        self,
        result_id: str,
        label_overrides: Optional[LabelOverrides] = None,
        result_cache_metadata: Optional[ResultCacheMetadata] = None,
        result_size_dimensions_limits: ResultSizeDimensions = (),
        result_size_bytes_limit: Optional[int] = None,
        use_local_ids_in_headers: bool = False,
    ) -> Tuple[pandas.DataFrame, DataFrameMetadata]:
        """
        Creates a data frame using an execution result's metadata identified by result_id. The data frame will respect
        the dimensionality specified in execution definition's result spec.

        Each dimension may be sliced by multiple labels. The factory will create MultiIndex for the dataframe's
        row index and the columns.

        Example of label_overrides structure:

        .. code-block:: python

            {
                "labels": {
                    "local_attribute_id": {
                        "title": "My new attribute label"
                    ,...
                },
                "metrics": {
                    "local_metric_id": {
                        "title": "My new metric label"
                    },...
                }
            }

        :param result_id: executionResult ID from ExecutionResponse
        :param label_overrides: label overrides for metrics and attributes
        :param result_cache_metadata: Metadata for the corresponding exec result. Optional.
        :param result_size_dimensions_limits: A tuple containing maximum size of result dimensions. Optional.
        :param result_size_bytes_limit: Maximum size of result in bytes. Optional.
        :param use_local_ids_in_headers: Use local identifiers of header attributes and metrics. Optional.

        :return: tuple holding DataFrame and DataFrame metadata
        """
        if label_overrides is None:
            label_overrides = {}

        if result_cache_metadata is None:
            result_cache_metadata = self.result_cache_metadata_for_exec_result_id(result_id=result_id)

        return convert_execution_response_to_dataframe(
            execution_response=BareExecutionResponse(
                api_client=self._sdk.client,
                workspace_id=self._workspace_id,
                execution_response=models.AfmExecutionResponse(
                    result_cache_metadata.execution_response, _check_type=False
                ),
            ),
            result_cache_metadata=result_cache_metadata,
            label_overrides=label_overrides,
            result_size_dimensions_limits=result_size_dimensions_limits,
            result_size_bytes_limit=result_size_bytes_limit,
            use_local_ids_in_headers=use_local_ids_in_headers,
        )



# (C) 2021 GoodData Corporation
from __future__ import annotations

from typing import Optional, Union

import pandas

from gooddata_pandas.data_access import compute_and_extract
from gooddata_pandas.utils import IndexDef, LabelItemDef, make_pandas_index
from gooddata_sdk import Attribute, Filter, GoodDataSdk, ObjId, SimpleMetric


class SeriesFactory:
    def __init__(self, sdk: GoodDataSdk, workspace_id: str) -> None:
        self._sdk = sdk
        self._workspace_id = workspace_id

    def indexed(
        self,
        index_by: IndexDef,
        data_by: Union[SimpleMetric, str, ObjId, Attribute],
        filter_by: Optional[Union[Filter, list[Filter]]] = None,
    ) -> pandas.Series:
        """
        Creates pandas Series from data points calculated from a single `data_by` that will be computed on granularity
        of the index labels. The elements of the index labels will be used to construct simple or hierarchical index.

        :param index_by: label to index by; specify either:

         - string with id: ``some_label_id``,
         - object identifier: ``ObjId(id='some_label_id', type='label')``,
         - string representation of object identifier: ``label/some_label_id``
         - or an Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - dict containing mapping of index name to label to use for indexing - specified in one of the ways list above

        :param data_by: label, fact or metric to that will provide data (metric values or label elements); specify
         either:

         - object identifier: ``ObjId(id='some_id', type='<type>')`` - where type is either ``label``, ``fact``
           or ``metric``
         - string representation of object identifier: ``<type>/some_id`` - where type is either ``label``, ``fact`` or
           ``metric``
         - Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - SimpleMetric object used in the compute model: ``SimpleMetric(local_id=..., item=..., aggregation=...)``

        :param filter_by: optionally specify filter to apply during computation on the server, reference to filtering
         column can be one of:

         - string reference to index key
         - object identifier in string form
         - object identifier: ``ObjId(id='some_label_id', type='<type>')``
         - Attribute or Metric depending on type of filter

        :return: pandas series instance

        """
        data, index = compute_and_extract(
            self._sdk,
            self._workspace_id,
            index_by=index_by,
            columns={"_series": data_by},
            filter_by=filter_by,
        )

        _idx = make_pandas_index(index)

        return pandas.Series(data=data["_series"], index=_idx)

    def not_indexed(
        self,
        data_by: Union[SimpleMetric, str, ObjId, Attribute],
        granularity: Optional[Union[list[LabelItemDef], IndexDef]] = None,
        filter_by: Optional[Union[Filter, list[Filter]]] = None,
    ) -> pandas.Series:
        """
        Creates pandas Series from data points calculated from a single `data_by` that will be computed on granularity
        of the specified labels. No index will be constructed.

        Note that data_by may also be a label in which case the Series will contain label elements.

        :param data_by: label, fact or metric to get data from; specify either:

         - object identifier: ``ObjId(id='some_id', type='<type>')`` - where type is either ``label``, ``fact``
           or ``metric``
         - string representation of object identifier: ``<type>/some_id`` - where type is either ``label``, ``fact`` or
           ``metric``
         - Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - SimpleMetric object used in the compute model: ``SimpleMetric(local_id=..., item=..., aggregation=...)``

        :param granularity: optionally specify label to slice the metric by; specify either:

         - string with id: ``some_label_id``,
         - object identifier: ``ObjId(id='some_label_id', type='label')``,
         - string representation of object identifier: ``label/some_label_id``
         - or an Attribute object used in the compute model: ``Attribute(local_id=..., label='some_label_id')``
         - list containing multiple labels to slice the metric by - specified in one of the ways list above
         - dict containing mapping of index name to label to use for indexing - specified in one of the ways list above;
           this option is available so that you can easily switch from indexed factory method to this one if needed

        :param filter_by: optionally specify filter to apply during computation on the server, reference to filtering
         column can be one of:

         - object identifier in string form
         - object identifier: ``ObjId(id='some_label_id', type='<type>')``
         - Attribute or Metric depending on type of filter

        :return: pandas series instance
        """

        if isinstance(granularity, list):
            _index: Optional[IndexDef] = {str(idx): label for idx, label in enumerate(granularity)}
        else:
            _index = granularity

        data, _ = compute_and_extract(
            self._sdk,
            self._workspace_id,
            index_by=_index,
            columns={"_series": data_by},
            filter_by=filter_by,
        )

        return pandas.Series(data=data["_series"])
